{-# LANGUAGE BangPatterns, NoMonomorphismRestriction #-}
module Main

where

import Control.Applicative
import Control.Monad
import Control.Monad.Random


import Data.List
import Data.Foldable ( for_ )
import Data.Maybe
import Data.Monoid
import qualified Data.IntSet as I

import Options.Applicative hiding ( Mod )

import System.Random


type Nat = Int  -- for documentation, not enforced


data Fm mod
 = PAt  Nat
 | NAt  Nat
 | Con  [Fm mod]
 | Dis  [Fm mod]
 | Dia  mod (Fm mod)
 | Box  mod (Fm mod)

data Query mod
 = [Fm mod] :|-: Fm mod


class Mod mod where
  showDia :: mod -> ShowS
  showBox :: mod -> ShowS

data K = K

instance Mod K where
  showDia _ = showString "<r>"
  showBox _ = showString "[r]"

newtype G = G Nat

instance Mod G where
  showDia (G n) = showString "{>= " . shows n . showString " r}"
  showBox (G n) = showString "{<= " . shows n . showString " r}"


newtype CL = CL I.IntSet

instance Mod CL where
  showDia (CL ags) = showString "<{" . sepWith " " (I.toAscList ags) . showString "}>"
  showBox (CL ags) = showString "[{" . sepWith " " (I.toAscList ags) . showString "}]"

instance Mod mod => Show (Fm mod) where
  showsPrec _ fm
   = case fm of
       PAt n    -> showString  "p" . shows n
       NAt n    -> showString "~p" . shows n

       Con []   -> showString "True"
       Con [f]  -> shows f
       Con fs   -> showParen True $ sepWith " & " fs

       Dis []   -> showString "False"
       Dis [f]  -> shows f
       Dis fs   -> showParen True $ sepWith " | " fs

       Dia m f -> showDia m . shows f
       Box m f -> showBox m . shows f

top,bot :: Fm mod
top = Con []
bot = Dis []

instance Mod mod => Show (Query mod) where
  showsPrec _ (tbox :|-: f)
   = sepWith "; " tbox . showString " |- " . shows f

sepWith :: Show a => String -> [a] -> ShowS
sepWith s = appEndo . mconcat . map Endo . intersperse (showString s) . map shows

data GenFmConf nat bool float
 = FmConf {
    depth  :: nat      -- modal depth of the resulting formula
   ,numAts :: nat      -- number of atoms
   ,minCon :: nat      -- min number of conjuncts per conjunction (0 allows for Top)
   ,maxCon :: nat      -- max number of conjuncts per conjunction
   ,minDis :: nat      -- min number of disjuncts per disjunction (0 allows for Bot)
   ,maxDis :: nat      -- max number of disjuncts per disjunction
   ,probAt :: float    -- probability of having an atom
   ,topCon :: bool     -- allow conjunctions at the top level
   ,topDis :: bool     -- allow disjunctions at the top level
   }

type FmConf    = GenFmConf Nat Bool Float
type CmdFmConf = GenFmConf (Last Nat) (Last Bool) (Last Float)


instance (Monoid a, Monoid b, Monoid c) => Monoid (GenFmConf a b c) where
  mempty
   = FmConf mempty mempty mempty mempty mempty mempty mempty mempty mempty

  mappend l r
   = FmConf {
       depth  = m depth
      ,numAts = m numAts
      ,minCon = m minCon
      ,maxCon = m maxCon
      ,minDis = m minDis
      ,maxDis = m maxDis
      ,probAt = m probAt
      ,topCon = m topCon
      ,topDis = m topDis
     }
     where
      m f
       = f l <> f r

overrideWith :: FmConf -> CmdFmConf -> FmConf
overrideWith b o
 = FmConf {
   depth  = override depth  depth
  ,numAts = override numAts numAts
  ,minCon = override minCon minCon
  ,maxCon = override maxCon maxCon
  ,minDis = override minDis minDis
  ,maxDis = override maxDis maxDis
  ,probAt = override probAt probAt
  ,topDis = override topDis topDis
  ,topCon = override topCon topCon
  }
  where
    override f g
     = fromMaybe (f b) (getLast $ g o)

data QueryConf
 = QueryConf {
    formConf :: FmConf
   ,tboxSize :: Nat
   ,tboxConf :: FmConf
   }


-- Modality generators
monoK :: Rand K
monoK = pure K


gradesBetween :: Nat -> Nat -> Rand G
gradesBetween from to = G <$> inRange (from,to)



numAgents :: Nat -> Rand CL
numAgents n = CL <$> buildSet I.empty n
  where
    buildSet !acc m
      | m <= 0    = pure acc
      | otherwise = do heads <- rand
                       let acc' = if heads then acc else I.insert m acc
                       buildSet acc' (m-1)

someFormula :: Rand mod -> FmConf -> Rand (Fm mod)
someFormula modGen conf = go (depth conf)
  where
    lits
     = let as = [1..1+numAts conf] in [top,bot] ++ map PAt as ++ map NAt as

    juncts junct_type r = \f ->
       do n <- inRange r
          junct_type <$> exactly n f

    formulaOfDepth d
     = do (diabox,xxjuncts) <- oneOf [(Dia,Con `juncts` (minCon conf,maxCon conf))
                                     ,(Box,Dis `juncts` (minDis conf,maxDis conf))]
          diabox <$> modGen <*> xxjuncts (go d)

    p = max 0 $ min 1 $ probAt conf

    go d
     | d == 0
      = oneOf lits

     | d == depth conf
      = let modal_form = formulaOfDepth (d-1)
        in join $ oneOf $ concat
            [ [modal_form]
            , [Con `juncts` (minCon conf,maxCon conf) $ modal_form | topCon conf]
            , [Dis `juncts` (minDis conf,maxDis conf) $ modal_form | topDis conf]
            ]

     | otherwise
      = join $ fromFreqs [oneOf lits           `withFreq`    p
                         ,formulaOfDepth (d-1) `withFreq` (1-p)]

someQuery :: Rand mod -> QueryConf -> Rand (Query mod)
someQuery modGen QueryConf{tboxSize = size, tboxConf = tconf, formConf = fconf}
 = (:|-:) <$> exactly size (someFm tconf) <*> someFm fconf
 where
   someFm = someFormula modGen

exactly :: Nat -> Rand a -> Rand [a]
exactly n = sequence . take n . repeat

main :: IO ()
main
 = do (randQuery,n,m_seed) <- execParser opts_info

      for_ m_seed $ \given_seed ->
        setStdGen (mkStdGen given_seed)

      sequence_ $ take n $ repeat (putStrLn =<< pick randQuery)

 where
  opts_info
   = info (helper <*> (infoOption "0.1" ( long "version" ) <*> opts)) mempty

  opts
   = (,,) <$> (genSel <*> queryConf) <*> numQueries <*> optional seed


genSel :: Parser (QueryConf -> Rand String)
genSel
 = (gen <$> selK) <|> (gen <$> selG) <|> (gen <$> selCL)

 where
  gen :: Mod mod => Rand mod -> QueryConf -> Rand String
  gen modGen conf = show <$> someQuery modGen conf

  selK
   = flag' monoK
       (  long "monomodal-K"
       <> short 'K'
       <> help "Language is K with one modality"
       )

  selG
   = flag' gradesBetween
       (  long "monomodal-gml"
       <> short 'G'
       <> help "Language is GML with one modality"
       )
     <*>
     option (  value 0
            <> long "min-grade"
            <> metavar "NUM"
            <> help "Minimum possible grade"
            <> showDefault
            <> hidden
            )
     <*>
     option (  value 7
            <> long "max-grade"
            <> metavar "NUM"
            <> help "Maximum possible grade"
            <> showDefault
            <> hidden
            )

  selCL
   = flag' numAgents
       (  long "coalition-logic"
       <> help "Language is Coalition Logic"
       )
     <*>
     option (  value 5
            <> long "num-agents"
            <> short 'A'
            <> metavar "NUM"
            <> help "Number of agents"
            <> showDefault
            <> hidden
            )

queryConf :: Parser QueryConf
queryConf
 = build
    <$> fmConf TgBoth
    <*> fmConf TgForm
    <*> fmConf TgTBox
    <*> option (  value 0
               <> long "tbox-formulas"
               <> short 'T'
               <> metavar "NUM"
               <> help "Number of formulas to include on the TBox"
               <> showDefault
               )
 where
   build commonConf formOnly tboxOnly size
    = QueryConf {
       tboxSize = size
      ,tboxConf = defaultFmConf `overrideWith` (commonConf <> tboxOnly)
      ,formConf = defaultFmConf `overrideWith` (commonConf <> formOnly)
      }

data FmConfTarget
 = TgForm | TgTBox | TgBoth
  deriving Eq

defaultFmConf :: FmConf
defaultFmConf
 = FmConf {
    depth  = 3
   ,numAts = 5
   ,minCon = 2
   ,maxCon = 6
   ,minDis = 2
   ,maxDis = 6
   ,probAt = 0.25
   ,topCon = True
   ,topDis = True
   }

fmConf :: FmConfTarget -> Parser CmdFmConf
fmConf tg
 = FmConf
    <$> _option  depth  "modal-depth"   "Modal depth of the formulas"   (metavar "NUM" <> _short 'd')
    <*> _option  numAts "propositions"  "Number of proposition symbols" (metavar "NUM" <> _short 'p')
    <*> _option  minCon "min-conjuncts" "Minimum number of conjuncts per conjunction" (metavar "NUM")
    <*> _option  maxCon "max-conjuncts" "Maximum number of conjuncts per conjunction" (metavar "NUM")
    <*> _option  minDis "min-disjuncts" "Minimum number of disjuncts per disjunction" (metavar "NUM")
    <*> _option  maxDis "max-disjuncts" "Maximum number of disjuncts per disjunction" (metavar "NUM")
    <*> _option  probAt "prob-atom"     "Probability of an atom occurring instead of a modality" (metavar "[0..1]")
    <*> _turnOff "no-toplevel-conj"     "Don't generate conjunctions at the top-level" mempty
    <*> _turnOff "no-toplevel-disj"     "Don't generate disjunctions at the top-level" mempty
 where
  _option field s h mods
   = Last <$> (optional $ option (long (mkLong s) <> hidden <> help (mkDescr (Just $ show . field) h) <> mods))

  _turnOff s h mods
   = Last <$> (optional $ flag' False (long (mkLong s) <> hidden <> help (mkDescr Nothing h) <> mods))

  mkLong
   = case tg of
       TgForm -> (++ "-form")
       TgTBox -> (++ "-tbox")
       TgBoth -> id

  mkDescr mf
   = (++ ' ':msg)
   where
    msg
     = case tg of
         TgForm -> "(override only for the formula)"
         TgTBox -> "(override only for TBox formulas)"
         TgBoth -> maybe "" (\f -> "(default " ++ (f defaultFmConf) ++ ")") mf

  _short
   = if tg == TgBoth then short else const mempty

numQueries :: Parser Nat
numQueries
 = option (  value 1
          <> long "num-queries"
          <> short 'n'
          <> metavar "NUM"
          <> help "Number of queries to generate"
          <> showDefault
          )

seed :: Parser Int
seed
 = option (  long "seed"
          <> metavar "NUM"
          <> help "Random seed"
          )

