(** A graph-tableau-based decision procedure framework for coalgebraic logics.
    @author Florian Widmann
 *)


open CoAlgMisc

module M = Minisat

type sortTable = CoAlgMisc.sortTable

type nomTable = string -> CoAlgFormula.sort option

type assumptions = CoAlgFormula.sortedFormula list

type input = sortTable * nomTable * assumptions * CoAlgFormula.sortedFormula


(*****************************************************************************)
(*                     Propagation of Satisfiability                         *)
(*****************************************************************************)

let propSatFindSucc setCnstr cset =
  if csetHasDot cset then false
  else
    match graphFindCnstr cset with
      | None -> assert false
      | Some SatC -> true
      | Some (OpenC _) -> setMemCnstr setCnstr cset
      | Some (UnexpandedC _)
      | Some UnsatC -> false

let rec propSat setStates setCores setCnstr = function
  | [] -> ()
  | propElem::tl ->
      let tl1 =
        match propElem with
        | UState (state, _) ->
            if setMemState setStates state then
              let cnstrs = stateGetConstraints state in
              let isStillOpen = cssExists (fun c -> propSatFindSucc setCnstr c) cnstrs in
              if isStillOpen then () else setRemoveState setStates state
            else ();
            tl
        | UCore (core, _) ->
            if setMemCore setCores core then
              let cnstrs = coreGetConstraints core in
              let isStillOpen = cssExists (fun c -> propSatFindSucc setCnstr c) cnstrs in
              if isStillOpen then tl else begin
                setRemoveCore setCores core;
                let cnstrPar = coreGetConstraintParents core in
                List.fold_left (fun acc cset -> (UCnstr cset)::acc) tl cnstrPar
              end
            else tl
        | UCnstr cset ->
            if setMemCnstr setCnstr cset then begin
              setRemoveCnstr setCnstr cset;
              match graphFindCnstr cset with
              | Some (OpenC nodes) ->
                  let prop acc node =
                    match node with
                    | Core core -> (UCore (core, true))::acc
                    | State state -> (UState (state, None))::acc
                  in
                  List.fold_left prop tl nodes
              | _ -> assert false
            end else tl
      in
      propSat setStates setCores setCnstr tl1

let propagateSat () =
  let setStates = setEmptyState () in
  let setCores = setEmptyCore () in
  let setCnstr = setEmptyCnstr () in
  let fktS1 state =
    match stateGetStatus state with
    | Expandable
    | Open -> setAddState setStates state
    | Unsat
    | Sat -> ()
  in
  graphIterStates fktS1;
  let fktC1 core =
    match coreGetStatus core with
    | Expandable
    | Open -> setAddCore setCores core
    | Unsat
    | Sat -> ()
  in
  graphIterCores fktC1;
  let cfkt1 cset cnstr =
    if csetHasDot cset then ()
    else
      match cnstr with
      | OpenC _ -> setAddCnstr setCnstr cset
      | UnsatC
      | SatC
      | UnexpandedC _ -> ()
  in
  graphIterCnstrs cfkt1;
  graphIterStates (fun state -> propSat setStates setCores setCnstr [UState (state, None)]);
  graphIterCores (fun core -> propSat setStates setCores setCnstr [UCore (core, false)]);
  setIterState (fun state -> stateSetStatus state Sat) setStates;
  let setCoreSat core =
    coreSetStatus core Sat;
    if core == graphGetRoot () then raise (CoAlg_finished true) else ()
  in
  setIterCore setCoreSat setCores;
  setIterCnstr (fun cset -> graphReplaceCnstr cset SatC) setCnstr


(*****************************************************************************)
(*                     Propagation of Unsatisfiability                       *)
(*****************************************************************************)

let isConstraintUnsat cset =
  match graphFindCnstr cset with
  | None -> assert false
  | Some UnsatC -> true
  | _ -> false

let fhtMustFind fht f =
  match fhtFind fht f with
  | Some l -> l
  | None -> assert false

let lhtMustFind lht l =
  match lhtFind lht l with
  | Some f -> f
  | None -> assert false

let rec propagateUnsat = function
  | [] -> ()
  | propElem::tl ->
      let tl1 =
        match propElem with
        | UState (state, idxopt) -> begin
            match stateGetStatus state with
            | Unsat -> tl
            | Sat -> assert false
            | Open
            | Expandable ->
                let mbs =
                  match idxopt with
                  | None -> stateGetBs state
                  | Some idx ->
                      let (dep, children) = stateGetRule state idx in
                      let getBs core =
                        assert (coreGetStatus core = Unsat);
                        coreGetBs core
                      in
                      dep (List.map getBs children)
                in
                stateSetBs state mbs;
                stateSetStatus state Unsat;
                let prop acc core =
                  let turnsUnsat =
                    match coreGetStatus core with
                    | Open -> List.for_all (fun s -> stateGetStatus s = Unsat) (coreGetChildren core)
                    | Expandable
                    | Unsat
                    | Sat -> false
                  in
                  if turnsUnsat then (UCore (core, false))::acc else acc
                in
                List.fold_left prop tl (stateGetParents state)
          end
        | UCore (core, comesFromCnstr) -> begin
            match coreGetStatus core with
            | Unsat -> tl
            | Sat -> assert false
            | Open
            | Expandable ->
                let mbs =
                  if comesFromCnstr then coreGetBs core
                  else
                    let bs = coreGetBs core in
                    let solver = coreGetSolver core in
                    let fht = coreGetFht core in
                    let lht = lhtInit () in
                    let addLits f acc =
                      let lit = fhtMustFind fht f in
                      lhtAdd lht lit f;
                      lit::acc
                    in
                    let litset = bsetFold addLits bs [] in
                    let addClauses state =
                      let cbs = stateGetBs state in
                      let clause = bsetFold (fun f acc -> (M.neg_lit (fhtMustFind fht f))::acc) cbs [] in
                      let okay = M.add_clause solver clause in
                      assert okay
                    in
                    List.iter addClauses (coreGetChildren core);
                    let isSat = M.invoke_solver solver litset in
                    assert (not isSat);
                    let res = bsetMake () in
                    let confls = M.get_conflict solver in
                    List.iter (fun l -> bsetAdd res (lhtMustFind lht l)) confls;
                    res
                in
                coreSetBs core mbs;
                coreSetStatus core Unsat;
                if core == graphGetRoot () then raise (CoAlg_finished false) else ();
                let prop acc (state, idx) =
                  let turnsUnsat =
                    match stateGetStatus state with
                    | Open
                    | Expandable ->
                        List.for_all (fun c -> coreGetStatus c = Unsat) (snd (stateGetRule state idx))
                    | Unsat
                    | Sat -> false
                  in
                  if turnsUnsat then (UState (state, Some idx))::acc else acc
                in
                let tl2 = List.fold_left prop tl (coreGetParents core) in
                List.fold_left (fun acc cset -> (UCnstr cset)::acc) tl2 (coreGetConstraintParents core)
          end
        | UCnstr cset ->
            match graphFindCnstr cset with
            | None
            | Some SatC -> assert false
            | Some UnsatC -> tl
            | Some (UnexpandedC nodes)
            | Some (OpenC nodes) ->
                graphReplaceCnstr cset UnsatC;
                let prop acc node =
                  match node with
                  | State state ->
                      let turnsUnsat = 
                        match stateGetStatus state with
                        | Expandable
                        | Open -> cssForall isConstraintUnsat (stateGetConstraints state)
                        | Unsat
                        | Sat -> false
                      in
                      if turnsUnsat then (UState (state, None))::acc else acc
                  | Core core ->
                      let turnsUnsat =
                        match coreGetStatus core with
                        | Expandable
                        | Open -> cssForall isConstraintUnsat (coreGetConstraints core)
                        | Unsat
                        | Sat -> false
                      in
                      if turnsUnsat then (UCore (core, true))::acc else acc
                in
                List.fold_left prop tl nodes
      in
      propagateUnsat tl1


(*****************************************************************************)
(*                   Propagation of Nominal Constraints                      *)
(*****************************************************************************)

let extractAts sort bs =
  let ats = csetMake () in
  let procNom nom f =
    match lfGetType sort f with
    | AndF
    | OrF -> ()
    | TrueFalseF
    | AtF -> ()
    | _ -> csetAdd ats (lfGetAt (sort, nom) f)
  in
  let getAts f =
    match lfGetType sort f with
    | AtF -> csetAdd ats (lfToAt sort f)
    | NomF -> bsetIter (fun f1 -> procNom f f1) bs
    | _ -> ()
  in
  bsetIter getAts bs;
  ats

let detConstraintsState state =
  let procRule accR (_, chldrn) =
    let procChldrn accC core =
      if coreGetStatus core = Unsat then accC
      else cssUnion accC (coreGetConstraints core)
    in
    let orset = List.fold_left procChldrn cssEmpty chldrn in
    let procOrset csetO accO =
      let mkUnion csetU accU =
        let cset = csetUnion csetO csetU in
        cssAdd cset accU
      in
      cssFold mkUnion accR accO
    in
    cssFold procOrset orset cssEmpty
  in
  let sort = stateGetSort state in
  let bs = stateGetBs state in
  let ats = extractAts sort bs in
  if stateGetStatus state = Expandable then csetAddDot ats else ();
  List.fold_left procRule (cssSingleton ats) (stateGetRules state)

let detConstraintsCore core =
  let sort = coreGetSort core in
  let bs = coreGetBs core in
  let ats = extractAts sort bs in
  let addCnstr acc state =
    if stateGetStatus state = Unsat then acc
    else
      let csets = stateGetConstraints state in
      (* cssFold (fun c a -> cssAdd (csetUnion c ats) a) csets acc *)
      cssUnion csets acc
  in
  let initInner =
    if coreGetStatus core = Expandable then
      (* let cset = csetCopy ats in *)
      let cset = ats in
      csetAddDot cset;
      cssSingleton cset
    else cssEmpty
  in
  List.fold_left addCnstr initInner (coreGetChildren core)

let rec propNom = function
  | [] -> ()
  | node::tl ->
      let tl1 =
        match node with
        | State state ->
            if stateGetStatus state = Unsat then tl
            else
              let csets = detConstraintsState state in
              let oldcsets = stateGetConstraints state in
              if cssEqual csets oldcsets then tl
              else begin
                stateSetConstraints state csets;
                List.fold_left (fun acc c -> (Core c)::acc) tl (stateGetParents state)
              end
        | Core core ->
            if coreGetStatus core = Unsat then tl
            else
              let csets = detConstraintsCore core in
              let oldcsets = coreGetConstraints core in
              if cssEqual csets oldcsets then tl
              else begin
                coreSetConstraints core csets;
                List.fold_left (fun acc (s, _) -> (State s)::acc) tl (coreGetParents core)
              end
      in
      propNom tl1

let updateConstraints node elemU csets =
  let update cset acc =
    let cnstr =
      match graphFindCnstr cset with
      | Some cn -> cn
      | None ->
          let cn = UnexpandedC [] in
          graphInsertCnstr cset cn;
          queueInsertCnstr cset;
          cn
    in
    match cnstr with
    | UnsatC -> acc
    | UnexpandedC parents ->
        graphReplaceCnstr cset (UnexpandedC (node::parents));
        false
    | OpenC parents ->
        graphReplaceCnstr cset (OpenC (node::parents));
        false
    | SatC -> false
  in
  let isUnsat = cssFold update csets true in
  if isUnsat then propagateUnsat [elemU] else ()

let propagateNominals () =
  let init = cssSingleton (csetMake ()) in
  graphIterStates (fun s -> if stateGetStatus s = Unsat then () else stateSetConstraints s init);
  graphIterCores (fun c -> if coreGetStatus c = Unsat then () else coreSetConstraints c init);
  graphIterStates (fun s -> propNom [State s]);
  graphIterCores (fun c -> propNom [Core c]);
  graphClearCnstr ();
  let fktS state =
    if stateGetStatus state = Unsat then ()
    else updateConstraints (State state) (UState (state, None)) (stateGetConstraints state)
  in
  graphIterStates fktS;
  let fktC core =
    if coreGetStatus core = Unsat then ()
    else updateConstraints (Core core) (UCore (core, true)) (coreGetConstraints core)
  in
  graphIterCores fktC


(*****************************************************************************)
(*                           Node Expansions                                 *)
(*****************************************************************************)

let getLit sort fht solver f =
  match fhtFind fht f with
  | Some lit -> lit
  | None ->
      let var = M.new_variable solver in
      let lit = M.var_to_lit var true in
      fhtAdd fht f lit;
      let () =
        match lfGetNeg sort f with
        | None -> ()
        | Some nf ->
            let nlit = M.neg_lit lit in
            fhtAdd fht nf nlit;
      in
      lit

let newCore sort bs =
  let fht = fhtInit () in
  let solver = M.new_solver () in
  let rec addClauses f =
    let lf = getLit sort fht solver f in
    match lfGetType sort f with
    | OrF ->
        let f1 = lfGetDest1 sort f in
        let f2 = lfGetDest2 sort f in
        addClauses f1;
        addClauses f2;
        let lf1 = fhtMustFind fht f1 in
        let lf2 = fhtMustFind fht f2 in
        let okay = M.add_clause solver [M.neg_lit lf; lf1; lf2] in
        assert (okay)
    | AndF ->
        let f1 = lfGetDest1 sort f in
        let f2 = lfGetDest2 sort f in
        addClauses f1;
        addClauses f2;
        let lf1 = fhtMustFind fht f1 in
        let lf2 = fhtMustFind fht f2 in
        let okay1 = M.add_clause solver [M.neg_lit lf; lf1] in
        assert (okay1);
        let okay2 = M.add_clause solver [M.neg_lit lf; lf2] in
        assert (okay2)
    | _ -> ()
  in
  bsetIter addClauses bs;
  coreMake sort bs solver fht

let getNextState core =
  let bs = coreGetBs core in
  let fht = coreGetFht core in
  let litset = bsetFold (fun f acc -> (fhtMustFind fht f)::acc) bs [] in
  let solver = coreGetSolver core in
  let isSat = M.invoke_solver solver litset in
  if not isSat then None
  else
    let sort = coreGetSort core in
    let newbs = bsetMake () in
    let rec mkExclClause f acc =
      match lfGetType sort f with
      | OrF ->
          let f1 = lfGetDest1 sort f in
          let lf1 = fhtMustFind fht f1 in
          if M.literal_status solver lf1 = M.LTRUE then mkExclClause f1 acc
          else
            let f2 = lfGetDest2 sort f in
            let lf2 = fhtMustFind fht f2 in
            assert (M.literal_status solver lf2 = M.LTRUE);
            mkExclClause f2 acc
      | AndF ->
          let f1 = lfGetDest1 sort f in
          let lf1 = fhtMustFind fht f1 in
          assert (M.literal_status solver lf1 = M.LTRUE);
          let acc1 = mkExclClause f1 acc in
          let f2 = lfGetDest2 sort f in
          let lf2 = fhtMustFind fht f2 in
          assert (M.literal_status solver lf2 = M.LTRUE);
          mkExclClause f2 acc1
      | _ ->
          bsetAdd newbs f;
          (M.neg_lit (fhtMustFind fht f))::acc
    in
    let clause = bsetFold mkExclClause bs [] in
    let okay = M.add_clause solver clause in
    assert (okay);
    Some (sort, newbs)

let newState sort bs =
  let (func, sl) = !sortTable.(sort) in
  let producer = CoAlgLogics.getExpandingFunctionProducer func in
  let exp = producer sort bs sl in
  stateMake sort bs exp

let insertState parent sort bs =
  let child =
    match graphFindState sort bs with
    | None ->
        let s = newState sort bs in
        graphInsertState sort bs s;
        queueInsertState s;
        s
    | Some s -> s
  in
  coreAddChild parent child;
  stateAddParent child parent

let expandCore core =
  match getNextState core with
  | Some (sort, bs) ->
      insertState core sort bs;
      queueInsertCore core
  | None ->
      let isUnsat = List.for_all (fun s -> stateGetStatus s = Unsat) (coreGetChildren core) in
      if isUnsat then propagateUnsat [UCore (core, false)]
      else coreSetStatus core Open


let insertCore sort bs =
  match graphFindCore sort bs with
  | None ->
      let c = newCore sort bs in
      graphInsertCore sort bs c;
      queueInsertCore c;
      c
  | Some c -> c

let insertRule state dep chldrn =
  let chldrn = listFromLazylist chldrn in
  let insert (isUns, acc) (sort, bs) =
    let bs1 = bsetAddTBox sort bs in
    let core = insertCore sort bs1 in
    let isUns1 = if coreGetStatus core = Unsat then isUns else false in
    (isUns1, core::acc)
  in
  let (isUnsat, children) = List.fold_left insert (true, []) chldrn in
  let idx = stateAddRule state dep (List.rev children) in
  List.iter (fun c -> coreAddParent c state idx) children;
  if isUnsat then begin
    propagateUnsat [UState (state, Some idx)];
    false
  end else true

let rec insertAllRules state = function
  | [] -> true
  | (dep, chldrn)::tl ->
      let notUnsat = insertRule state dep chldrn in
      if notUnsat then insertAllRules state tl else false

let expandState state =
  match stateNextRule state with
  | MultipleElements rules ->
      let notUnsat = insertAllRules state rules in
      if notUnsat then stateSetStatus state Open
  | SingleElement (dep, chldrn) ->
      let notUnsat = insertRule state dep chldrn in
      if notUnsat then queueInsertState state else ()
  | NoMoreElements -> stateSetStatus state Open


let expandCnstr cset =
  let nht = nhtInit () in
  let mkCores f =
    let (sort, lf) as nom = atFormulaGetNominal f in
    let nomset =
      match nhtFind nht nom with
      | Some ns -> ns
      | None ->
          let cset1 = csetCopy cset in
          csetRemDot cset1;
          let bs = csetAddTBox sort cset1 in
          bsetAdd bs lf;
          nhtAdd nht nom bs;
          bs
    in
    bsetAdd nomset (atFormulaGetSubformula f)
  in
  csetIter cset mkCores;
  let inCores (sort, _) bs (isUns, acc) =
    let core = insertCore sort bs in
    coreAddConstraintParent core cset;
    (coreGetStatus core = Unsat || isUns, core::acc)
  in
  let (isUnsat, children) = nhtFold inCores nht (false, []) in
  if isUnsat then propagateUnsat [UCnstr cset]
  else
    match graphFindCnstr cset with
    | Some (UnexpandedC parents) -> graphReplaceCnstr cset (OpenC parents)
    | _ -> assert false


(*****************************************************************************)
(*                           The Main Loop                                   *)
(*****************************************************************************)

let rec expandNodesLoop () =
  match queueGetElement () with
  | QState state ->
      if stateGetStatus state = Expandable then begin
        expandState state;
        if doNominalPropagation () then begin
          propagateNominals ();
          if doSatPropagation () then propagateSat ()
        end else ()
      end else ();
      expandNodesLoop ()
  | QCore core ->
      if coreGetStatus core = Expandable then begin
        expandCore core;
        if doNominalPropagation () then begin
          propagateNominals ();
          if doSatPropagation () then propagateSat ()
        end else ()
      end else ();
      expandNodesLoop ()
  | QCnstr cset ->
      expandCnstr cset;
      expandNodesLoop ()
  | QEmpty -> ()

let rec buildGraphLoop () =
  expandNodesLoop ();
  propagateNominals ();
  if queueIsEmpty () then () else buildGraphLoop ()
    
(** A graph-tableau-based decision procedure framework for coalgebraic logics.
    @param verbose An optional switch which determines
    whether the procedure shall print some information on the standard output.
    The default is false.
    @param sorts An array mapping each sort (represented as an integer)
    to a functor and a list of sorts which are the arguments of the functor.
    @param nomTable A partial function mapping nominals (represented as strings)
    to their sorts.
    @param tbox A list of sorted formulae.
    @param sf A sorted formula.
    @return True if sf is satisfiable wrt tbox, false otherwise.
 *)
let isSat ?(verbose = false) sorts nomTable tbox sf =
  let start = if verbose then Unix.gettimeofday () else 0. in
  sortTable := Array.copy sorts;
  let (tbox1, sf1, bs) = ppFormulae nomTable tbox sf in
  let sort = fst sf in
  let bs1 = bsetAddTBox sort bs in
  graphInit ();
  queueInit ();
  let root = insertCore sort bs1 in
  graphAddRoot root;
  let sat =
    try
      buildGraphLoop ();
      match coreGetStatus (graphGetRoot ()) with
      | Expandable -> assert false
      | Unsat -> false
      | Sat
      | Open -> true
    with CoAlg_finished res -> res
  in
  if verbose then
    let stop = Unix.gettimeofday () in
    let addup lst = List.fold_left (fun acc sf -> acc + (CoAlgFormula.sizeSortedFormula sf)) 0 lst in
    print_newline ();
    print_endline ("Query: " ^ (CoAlgFormula.exportQuery tbox sf));
    let size = (CoAlgFormula.sizeSortedFormula sf) + (addup tbox) in
    print_endline ("Added Size: " ^ (string_of_int size));
    print_endline ("Negation Normal Form: " ^ (CoAlgFormula.exportQuery tbox1 sf1));
    let nsize = (CoAlgFormula.sizeSortedFormula sf1) + (addup tbox1) in
    print_endline ("Added Size: " ^ (string_of_int nsize));
    print_endline ("Result: Query is " ^ (if not sat then "not " else "") ^ "satisfiable.");
    print_endline ("Time: " ^ (string_of_float (stop -. start)));
    print_endline ("Generated states: " ^ (string_of_int (graphSizeState ())));
    print_endline ("Generated cores: " ^ (string_of_int (graphSizeCore ())));
    print_endline ("Generated constraints: " ^ (string_of_int (graphSizeCnstr ())));
    print_newline ()
  else ();
  sat
