
open Readline

(* name, callback, shortdescription, long description *)
type binding = string * (string list -> unit) * string * string

type settings = {
    prompt: string;
    bindings: binding list;
}

let session : settings option ref = ref (None)

let bind a b c d = (a,b,c,d)

(* (* for debugging purposes... *)
let cmp a b =
    print_endline ("First string: >"^a^"<");
    print_endline ("First string: >"^b^"<");
    print_endline ("a=b -> >"^(string_of_bool (a = b))^"<");
    print_endline ("a==b -> >"^(string_of_bool (a == b))^"<");
    print_endline ("a<>b -> >"^(string_of_bool (a <> b))^"<")
*)

(* find a command and execute the action for its binding *)
(* if the command does not exist, print some error message *)

let doForCommand (bindings: binding list) (commandName: string)
                 (action: binding -> unit) : unit =
    let findBinding (name: string) : binding =
        let checkname (n,_,_,_) = (name = n) in
        List.find checkname (bindings)
    in
    try
      let binding = findBinding commandName in
      action binding (* call the action *)
    with Not_found -> begin
      let msg = "Command »"^commandName^"« not found." in
      print_endline msg
    end

let helpCommand bindings args =
    (*
    let maxCommandLen = ref 0 in
    List.iter (fun (n,_) => maxCommandLen := Pervasives.max (Str.length n) !maxCommandLen) bindings
    *)
    match args with
    | [] ->
        (* in case of no arguments, just print all with shortdescription *)
        let printBinding (name,_, shortdesc, _) =
            Printf.printf "    %-10s %s\n" name shortdesc
        in
        List.iter printBinding bindings;
        flush stdout
    | (command::_) ->
        (* if a command name is given, print its long description *)
        doForCommand bindings command (fun (_,_,short,long) ->
                print_endline (command ^": "^short);
                if long <> ""
                then begin
                        print_endline "";
                        print_endline (long)
                end else ()
            )

let addHelpBinding bindings =
    (* do some workaround for this kind of recursive structure *)
    let rec newBinds withFunction =
        let d = "Prints this help" in
        List.append (if withFunction
        then [ bind "help" (helpCommand (newBinds false)) d "" ]
        else [ bind "help" (fun _ -> ()) d "" ]
        ) bindings
    in
    newBinds true

let start (settings: settings): unit =
    session := Some (settings);
    let rec loop () =
        if (!session = None) then ()
        else begin
            match (readline_prompt settings.prompt) with
            | None -> ()
            | Some line -> begin
                add_history line;
                let cmd = Str.split (Str.regexp "[ \n\t]+") line in
                (match cmd with
                | [] -> ()
                | (v::_) when v = "" -> ()
                | (name:: args) ->
                  doForCommand settings.bindings name (fun (_,callback,_,_) -> callback args)
                )
            end;
            loop()
        end
    in
    loop()

let stop () =
    session := None

let exitBinding : binding =
    bind "exit" (fun _ -> stop()) "exits this tool" ""

